/*
 * Author:  Cojocaru Adrian Viorel 332CA
 * Project: Proiect PM - Snake
 * File:    snake.c
 */

#include <avr/io.h>

#define F_CPU 16000000

#include <util/delay.h>
#include <stdlib.h>

#include "u8g.h"

// game defines
#define LCD_WIDTH  128
#define LCD_HEIGHT 64

#define UP    0
#define DOWN  1
#define LEFT  2
#define RIGHT 3
#define NONE  5

#define INITIAL_SIZE 50
#define MAX_SIZE     300

#define BLOCK_SIZE 2

#define HEAD 1
#define BODY 2
#define TAIL 3
#define FOOD 4

typedef struct {
    uint8_t x;
    uint8_t y;
} coord;

// directia de miscare (initial spre dreapta)
uint8_t snake_direction;
// matricea ecranului
coord snake[MAX_SIZE];
//
coord food;
// pozitia in vector a capului
uint32_t head_pos;
// pozitia in vector a capului
uint32_t tail_pos;
// dimensiunea sarpelui
uint32_t snake_size;

// LCD
u8g_t u8g;


/* Setare directii porturi folosite */
void init_buttons(void) {

    DDRA = 0xff;
    PORTA = 0x00;

    // UP
    DDRA &= ~(1 << PA0);
    PORTA |= (1 << PA0);
    
    // LEFT
    DDRA &= ~(1 << PA1);
    PORTA |= (1 << PA1);
    
    // RIGHT
    DDRA &= ~(1 << PA2);
    PORTA |= (1 << PA2);
    
    // DOWN
    DDRA &= ~(1 << PA3);
    PORTA |= (1 << PA3);
    
    // LED user
    DDRD |= (1 << PD7);
}

/* Verificare butoane apasate */
void check_input(void) {

    // UP
    if (!(PINA & (1 << PA0))) {
        snake_direction = UP;
    }
    
    // LEFT
    if (!(PINA & (1 << PA1))) {
        snake_direction = LEFT;
    }
    
    // RIGHT
    if (!(PINA & (1 << PA2))) {
        snake_direction = RIGHT;
    }
    
    // DOWN
    if (!(PINA & (1 << PA3))) {
        snake_direction = DOWN;
    }
}

void init_lcd(void) {
    MCUCR |= (1 << JTD); 
    MCUCR |= (1 << JTD);
    
    // init LCD - data ports PD0 - PD7, control on PC4 - PC1
    u8g_Init8Bit(&u8g, &u8g_dev_st7920_128x64_4x_8bit,
            PN(3, 0), PN(3, 1), PN(3, 2), PN(3, 3), 
            PN(3, 4), PN(3, 5), PN(3, 6), PN(3, 7), // DATA PORTS
            PN(2, 1),    // ENABLE
            U8G_PIN_NONE,
            U8G_PIN_NONE,
            PN(2, 3),    // RS / DI
            PN(2, 2),    // RW
            PN(2, 4));   // RESET
}

coord generate_coord(void) {
    coord c;
    
    c.x = rand() % LCD_WIDTH;
    c.y = rand() % LCD_HEIGHT;
    
    return c;
}

void init_snake(uint32_t seed) {
    uint8_t i;
    
    snake_size = INITIAL_SIZE;
    
    for (i = 0; i < INITIAL_SIZE; ++i) {
        snake[i].x = (LCD_WIDTH - INITIAL_SIZE) / 2 + i * BLOCK_SIZE;
        snake[i].y = LCD_HEIGHT / 2;
    }

    head_pos = INITIAL_SIZE - 1;
    tail_pos = 0;


    srand(seed);
    
    // generate food
    food = generate_coord();
}

void update_game(void) {
    uint32_t prev;

    // update head
    prev = head_pos;
    head_pos = (head_pos + 1) % MAX_SIZE;
    
    snake[head_pos] = snake[prev];
    
    switch(snake_direction) {
        case UP:
            snake[head_pos].y = (snake[head_pos].y + BLOCK_SIZE) % LCD_HEIGHT;
            break;
        case DOWN:
            snake[head_pos].y = (snake[head_pos].y - BLOCK_SIZE) % LCD_HEIGHT;
            break;
        case LEFT:
            snake[head_pos].x = (snake[head_pos].x - BLOCK_SIZE) % LCD_WIDTH;
            break;
        case RIGHT:
            snake[head_pos].x = (snake[head_pos].x + BLOCK_SIZE) % LCD_WIDTH;
            break;
    }
    
    // check food
    if (abs(food.x - snake[head_pos].x) < BLOCK_SIZE && abs(food.y - snake[head_pos].y) < BLOCK_SIZE) {
        ++snake_size;
        food = generate_coord();
    } else {
        // update tail
        tail_pos = (tail_pos + 1) % MAX_SIZE;
    }
}

void draw_start(void) {
    u8g_SetDefaultForegroundColor(&u8g);
    u8g_SetFont(&u8g, u8g_font_helvB10);
    u8g_DrawStr(&u8g, 5, 25, "Start Game!");
}

void draw_game(void) {
    uint32_t i;
    coord block;
    
    u8g_SetDefaultForegroundColor(&u8g);
    for (i = 0; i < snake_size; ++i) {
        block = snake[(tail_pos + i) % MAX_SIZE];
        u8g_DrawBox(&u8g, block.x - BLOCK_SIZE / 2, block.y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE);
    }
    
    // draw food
    u8g_DrawBox(&u8g, food.x - BLOCK_SIZE / 2, food.y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE);
}

int main(void) {

    init_buttons();
    init_lcd();
    
    snake_direction = NONE;
    
    // asteptam pana la prima comanda pentru a porni jocul
    while(1) {
        check_input();
        
        // start game
        if (snake_direction != NONE) break;
        
        u8g_FirstPage(&u8g);
        do {
            draw_start();
        } while (u8g_NextPage(&u8g));    
        
        _delay_ms(100);
    }
    
    init_snake(42);

    while (1) {
        check_input();
        update_game();
        
        u8g_FirstPage(&u8g);
        do {
            draw_game();
        } while (u8g_NextPage(&u8g));    
        _delay_ms(100);
    }
    
    return 0;
}
